from pathlib import Path
from iinfer.app import common, options
import glob
import io

"""
def main():
    option = options.Options()
    outdir = Path("tests") / "genapp"
    common.rmdirs(outdir)
    common.mkdirs(outdir)

    opt_files = glob.glob(f"iinfer/tools/datas/*.json", recursive=False)
    buffer =  io.StringIO()
    buffer.write(BASE_TEST)
    index = 0
    for opt_file in opt_files:
        opt = common.loadopt(opt_file)
        opt_list = option.mk_opt_list(opt)
        cmd_line = TEMP_TEST.format(index=index, mode_key=opt["mode"], cmd_key=opt["cmd"], opt_list=' " \\\n\t\t  "'.join(opt_list[0]))
        buffer.write(cmd_line)
        index += 1

    testfile = outdir / f"test_gentest.py"
    with open(testfile, "w", encoding="utf-8") as f:
        f.write(buffer.getvalue())
"""

def main():
    option = options.Options()
    outdir = Path("tests") / "genapp"
    common.rmdirs(outdir)
    common.mkdirs(outdir)

    def _set(tv, opt_name, opt_map):
        if type(tv) == list:
            for t in tv:
                if t is None or type(t) == bool and not t:
                    continue
                opt_map["opt_list"].append(f"--{opt_name}")
                if type(t) == bool:
                    continue
                if isinstance(t, Path):
                    opt_map["opt_list"].append(str(t).replace("\\", "/"))
                else:
                    opt_map["opt_list"].append(str(t))
            return
        elif tv is None or not tv:
            return
        opt_map["opt_list"].append(f"--{opt_name}")
        if type(tv) == bool:
            return
        if isinstance(tv, Path):
            opt_map["opt_list"].append(str(tv).replace("\\", "/"))
        else:
            opt_map["opt_list"].append(str(tv))

    for mode_key in option.get_mode_keys():
        if mode_key == '':
            continue
        index = 0
        buffer =  io.StringIO()
        buffer.write(BASE_TEST)
        for cmd_key in option.get_cmd_keys(mode_key):
            choices = option.get_cmd_choices(mode_key, cmd_key)
            names = set()
            for choice in choices:
                if "test_true" in choice:
                    names |= set(choice["test_true"].keys())
                if "test_false" in choice:
                    names |= set(choice["test_false"].keys())
            names = list(names)
            if len(names) <= 0:
                continue
            opt_list_true = dict()
            opt_list_chk = dict()
            opt_list_false = dict()
            opt_list_stdin = dict()
            for name in names:
                opt_list_true[name] = dict(opt_list=[], jadge=True)
                opt_list_chk[name] = dict(opt_list=[], jadge=False)
                opt_list_false[name] = dict(opt_list=[], jadge=False)
                for choice in choices:
                    set_false = True
                    if "test_false" in choice and name in choice["test_false"]:
                        testval = choice["test_false"][name]
                        _set(testval, choice["opt"], opt_list_chk[name])
                        _set(testval, choice["opt"], opt_list_false[name])
                        set_false = False
                    if "test_true" in choice:
                        if name in choice["test_true"]:
                            testval = choice["test_true"][name]
                        else:
                            testval = choice["test_true"][list(choice["test_true"].keys())[0]]
                        _set(testval, choice["opt"], opt_list_true[name])
                        if set_false: _set(testval, choice["opt"], opt_list_false[name])
                    if "test_stdin" in choice:
                        if name in choice["test_stdin"]:
                            testval = choice["test_stdin"][name]
                        else:
                            testval = choice["test_stdin"][list(choice["test_stdin"].keys())[0]]
                        opt_list_stdin[name] = testval
                if len(opt_list_true[name]["opt_list"]) <= 0:
                    del opt_list_true[name]
                if len(opt_list_chk[name]["opt_list"]) <= 0:
                    del opt_list_chk[name]
                    del opt_list_false[name]

            test_assert = option.get_cmd_attr(mode_key, cmd_key, "test_assert")
            test_assert = test_assert if test_assert is not None else "assert 'success' {jadge} result.keys()"
            jointxt = '",\n           "'
            for name in names:
                temp = TEMP_TEST
                stdin_file = ""
                if name in opt_list_stdin and opt_list_stdin[name] is not None:
                    temp = TEMP_TEST_STDIN
                    stdin_file = opt_list_stdin[name]
                temp = temp.replace("{test_assert}", test_assert)
                if name in opt_list_true:
                    opt_list = ["-m", mode_key, "-c", cmd_key] + opt_list_true[name]["opt_list"]
                    opt_list = [str(x) for x in opt_list if str(x) != "--debug"] + ["--debug"]
                    cmd_line = temp.format(index=index, mode_key=mode_key, cmd_key=cmd_key,
                                                name=name, jadge="in", opt_list=jointxt.join(opt_list),
                                                stdin_file=stdin_file)
                    buffer.write(cmd_line)
                    index += 1

                if name in opt_list_false:
                    opt_list = ["-m", mode_key, "-c", cmd_key] + opt_list_false[name]["opt_list"]
                    opt_list = [str(x) for x in opt_list if str(x) != "--debug"] + ["--debug"]
                    cmd_line = temp.format(index=index, mode_key=mode_key, cmd_key=cmd_key,
                                                name=name, jadge="not in", opt_list=jointxt.join(opt_list),
                                                stdin_file=stdin_file)
                    buffer.write(cmd_line)
                    index += 1

        testfile = outdir / f"test_{mode_key}.py"
        with open(testfile, "w", encoding="utf-8") as f:
            f.write(buffer.getvalue())


TEMP_TEST = """
@pytest.mark.run(order={index})
def test_{index}_{mode_key}_{cmd_key}_{name}(capfd):
    cmd = ["{opt_list}"]

    _, result = app.IinferApp().main(args_list=cmd)
    out, err = capfd.readouterr()
    print(out)
    {test_assert}

"""

TEMP_TEST_STDIN = """
@pytest.mark.run(order={index})
def test_{index}_{mode_key}_{cmd_key}_{name}(capfd, monkeypatch):
    cmd = ["{opt_list}"]

    with open('{stdin_file}', 'br') as f:
        sys.stdin = f.buffer = f
        _, result = app.IinferApp().main(args_list=cmd)
        out, err = capfd.readouterr()
        print(out)
        {test_assert}

"""

BASE_TEST = """
from iinfer.app import app
from pathlib import Path
from unittest.mock import patch
import iinfer
import os
import pytest
import time
import shutil
import subprocess
import sys


@pytest.fixture(scope='module', autouse=True)
def fixture_server():
    python = Path(iinfer.__file__).parent.parent / '.venv' / 'Scripts' / 'python.exe'
    cmd = f"{python} -m iinfer -m server -c start --svname server".split(' ')
    proc1 = subprocess.Popen(cmd)
    #cmd = f"{python} -m iinfer -m server -c start --svname server1".split(' ')
    #proc2 = subprocess.Popen(cmd)
    #cmd = f"{python} -m iinfer -m server -c start --svname server2".split(' ')
    #proc3 = subprocess.Popen(cmd)
    shutil.rmtree("mmdetection", ignore_errors=True)
    shutil.rmtree("mmpretrain", ignore_errors=True)
    shutil.rmtree("mmsegmentation", ignore_errors=True)
    time.sleep(15)
    yield
    #cmd = f"{python} -m iinfer -m server -c stop --svname server2 --timeout 15".split(' ')
    #subprocess.run(cmd)
    #cmd = f"{python} -m iinfer -m server -c stop --svname server1 --timeout 15".split(' ')
    #subprocess.run(cmd)
    cmd = f"{python} -m iinfer -m server -c stop --svname server --timeout 15".split(' ')
    subprocess.run(cmd)

"""


if __name__ == "__main__":
    main()



from iinfer.app.injections import before_grayimg_injection
from typing import Dict, Any
import logging

class BeforeGrayimgInjection2(before_grayimg_injection.BeforeGrayimgInjection):
    """
    このクラスは推論を実行する前処理のインジェクションクラスです。
    """
    def __init__(self, config:Dict[str,Any], logger:logging.Logger):
        """
        このクラスのインスタンスを初期化します。
        継承時は、このコンストラクタを呼び出すようにしてください。
            super().__init__(logger)
        Args:
            config (Dict[str,Any]): 設定
            logger (logging.Logger): ロガー
        """
        super().__init__(config, logger)



tta_model = dict(
    type='DetTTAModel',
    tta_cfg=dict(nms=dict(type='nms', iou_threshold=0.65), max_per_img=100))

img_scales = [(640, 640), (320, 320), (960, 960)]
tta_pipeline = [
    dict(type='LoadImageFromFile', backend_args=None),
    dict(
        type='TestTimeAug',
        transforms=[
            [
                dict(type='Resize', scale=s, keep_ratio=True)
                for s in img_scales
            ],
            [
                # ``RandomFlip`` must be placed before ``Pad``, otherwise
                # bounding box coordinates after flipping cannot be
                # recovered correctly.
                dict(type='RandomFlip', prob=1.),
                dict(type='RandomFlip', prob=0.)
            ],
            [
                dict(
                    type='Pad',
                    pad_to_square=True,
                    pad_val=dict(img=(114.0, 114.0, 114.0))),
            ],
            [dict(type='LoadAnnotations', with_bbox=True)],
            [
                dict(
                    type='PackDetInputs',
                    meta_keys=('img_id', 'img_path', 'ori_shape', 'img_shape',
                               'scale_factor', 'flip', 'flip_direction'))
            ]
        ])
]



from iinfer.app.injections import after_det_filter_injection
from typing import Tuple, Dict, Any
import logging


class AfterDetFilterInjection2(after_det_filter_injection.AfterDetFilterInjection):
    def __init__(self, config:Dict[str,Any], logger:logging.Logger):
        super().__init__(config, logger)


from pathlib import Path
from PIL import Image
from iinfer.app import common, predict
from iinfer.app.commons import convert
from typing import List, Tuple, Union, Any
import logging


SITE = 'https://huggingface.co/docs/diffusers/api/pipelines/stable_diffusion/text2img'
IMAGE_WIDTH = 640
IMAGE_HEIGHT = 640
REQUIREd_MODEL_CONF = False
REQUIREd_MODEL_WEIGHT = False

class Diffusers_StableDiffusionPipeline_Txt2Img(predict.TorchPredict):
    def __init__(self, logger:logging.Logger) -> None:
        super().__init__(logger)

    def create_session(self, deploy_dir:Path, model_path:Union[Path,Any], model_conf_path:Path, model_provider:str, gpu_id:int=None):
        """
        推論セッションを作成する関数です。
        startコマンド実行時に呼び出されます。
        この関数内でAIモデルのロードを行い、推論準備を完了するようにしてください。
        戻り値の推論セッションの型は問いません。

        Args:
            deploy_dir (Path): デプロイディレクトリのパス
            model_path (Path|Any): モデルファイルのパス
            model_conf_path (Path): モデル設定ファイルのパス
            gpu_id (int, optional): GPU ID. Defaults to None.

        Returns:
            推論セッション
        """
        from diffusers import StableDiffusionXLPipeline, EulerAncestralDiscreteScheduler

        opt = common.loadopt(model_conf_path) if model_conf_path is not None else dict()
        self.num_inference_steps = common.getopt(opt, 'num_inference_steps', preval=10, withset=False)

        # https://huggingface.co/martyn/sdxl-turbo-mario-merge-top-rated/blob/main/topRatedTurboxlLCM_v10.safetensors
        pipeline = StableDiffusionXLPipeline.from_single_file(model_path, cache_dir=deploy_dir)
        pipeline.scheduler = EulerAncestralDiscreteScheduler.from_config(pipeline.scheduler.config)
        if gpu_id is not None and self.is_gpu_available(model_path, model_conf_path, gpu_id):
            pipeline.to("cuda")
        pipeline.load_lora_weights('ntc-ai/SDXL-LoRA-slider.anime', weight_name='anime.safetensors', adapter_name="anime")
        pipeline.set_adapters(["anime"], adapter_weights=[2.0])

        return pipeline

    def predict(self, model, img_width:int, img_height:int, input_data:Union[Image.Image, str], labels:List[str]=None, colors:List[Tuple[int]]=None, nodraw:bool=False):
        """
        予測を行う関数です。
        predictコマンドやcaptureコマンド実行時に呼び出されます。
        引数のinput_dataが画像の場合RGBですので、戻り値の出力画像もRGBにしてください。
        戻り値の推論結果のdictは、通常推論結果項目ごとに値(list)を設定します。

        Args:
            model: 推論セッション
            img_width (int): モデルのINPUTサイズ（input_dataが画像の場合は、画像の幅）
            img_height (int): モデルのINPUTサイズ（input_dataが画像の場合は、画像の高さ）
            input_data (Image | str): 推論するデータ（画像の場合RGB配列であること）
            labels (List[str], optional): クラスラベルのリスト. Defaults to None.
            colors (List[Tuple[int]], optional): ボックスの色のリスト. Defaults to None.
            nodraw (bool, optional): 描画フラグ. Defaults to False.

        Returns:
            Tuple[Dict[str, Any], Image]: 予測結果と出力画像(RGB)のタプル
        """
        import torch
        output_images = model(input_data, negative_prompt='nsfw', guidance_scale = 2,
                              height=img_height, width=img_width,
                              num_inference_steps=self.num_inference_steps, output_type='pil',
                              torch_dtype=torch.float16).images

        return dict(prompt=input_data), output_images[0]



from iinfer.app.commons import convert
from iinfer.app.postprocesses import det_clip
from PIL import Image
from typing import Dict, Tuple, Any
import logging
import numpy as np


class DetFaceStore(det_clip.DetClip):
    def __init__(self, logger:logging.Logger, face_threshold:float=0.0, image_type:str='capture', clip_margin:int=0):
        """
        face Detectionの推論結果となったbbox部分を個別の画像として切り出す後処理クラスです。
        
        Args:
            logger (logging.Logger): ロガー
            face_threshold (float): 出力する顔スコアの閾値
            image_type (str): 切り出した画像のファイル形式
            clip_margin (int): bboxの周囲に余白を設けるピクセル数
            output_dir (Path): 切り出した画像を保存するディレクトリ
        """
        super().__init__(logger, image_type, clip_margin)
        self.face_threshold = face_threshold


    def post(self, outputs:Dict[str, Any], output_image:Image.Image) -> Tuple[Dict[str, Any], Image.Image]:
        """
        後処理を行う関数です。
        なおoutputsは、以下のような構造を持つDict[str, Any]です。
        {
            'success': {
                'output_boxes': List[List[int]],
                'output_embeddings': List[np.ndarray],
                'output_embedding_dtypes': List[str],
                'output_embedding_shapes': List[Tuple[int]],
                'output_scores': List[float],
                'output_image_name': str
            }
        }

        Args:
            outputs (Dict[str, Any]): 推論結果
            output_image (Image.Image): 入力画像（RGB配列であること）

        Returns:
            Dict[str, Any]: 後処理結果
            Image: 後処理結果
        """
        if 'success' not in outputs or type(outputs['success']) != dict:
            raise Exception('Invalid outputs. outputs[\'success\'] must be dict.')
        data = outputs['success']
        #if 'output_ids' not in data:
        #    raise Exception('Invalid outputs. outputs[\'success\'][\'output_ids\'] must be set.')
        #output_ids = data['output_ids']
        if 'output_boxes' not in data:
            raise Exception('Invalid outputs. outputs[\'success\'][\'output_boxes\'] must be set.')
        output_boxes = data['output_boxes']
        if 'output_embeddings' not in data:
            raise Exception('Invalid outputs. outputs[\'success\'][\'output_embeddings\'] must be set.')
        output_embeddings = data['output_embeddings']
        if 'output_embedding_dtypes' not in data:
            raise Exception('Invalid outputs. outputs[\'success\'][\'output_embedding_dtypes\'] must be set.')
        output_embedding_dtypes = data['output_embedding_dtypes']
        if 'output_embedding_shapes' not in data:
            raise Exception('Invalid outputs. outputs[\'success\'][\'output_embedding_shapes\'] must be set.')
        output_embedding_shapes = data['output_embedding_shapes']
        if 'output_scores' not in data:
            raise Exception('Invalid outputs. outputs[\'success\'][\'output_scores\'] must be set.')
        output_scores = data['output_scores']
        
        # 切り出し
        result = []
        for i, box in enumerate(output_boxes):
            if output_scores[i] >= 0 and output_scores[i] <= self.face_threshold:
                continue
            y1, x1, y2, x2 = box
            x1 = max(0, np.floor(x1 + 0.5).astype(int) - self.clip_margin)
            y1 = max(0, np.floor(y1 + 0.5).astype(int) - self.clip_margin)
            x2 = min(output_image.width, np.floor(x2 + 0.5).astype(int) + self.clip_margin)
            y2 = min(output_image.height, np.floor(y2 + 0.5).astype(int) + self.clip_margin)

            cropped_image = output_image.crop((x1, y1, x2, y2))
            face_image_npy = convert.img2npy(cropped_image)

            img_b64 = None
            #face_name = f"{output_image_name}.{i}.{self.image_type}"
            if self.image_type == 'capture' or self.image_type is None:
                img_b64 = convert.npy2b64str(face_image_npy)
            else:
                img_byte = convert.img2byte(cropped_image, format=self.image_type)
                img_b64 = convert.bytes2b64str(img_byte)
            result.append(dict(face_label='', face_embedding=output_embeddings[i], face_embedding_dtype=output_embedding_dtypes[i], face_embedding_shape=output_embedding_shapes[i],
                               face_score=output_scores[i], face_box=[x1, y1, x2, y2], face_image_type=self.image_type, face_image_shape=face_image_npy.shape, face_image=img_b64))
            #result.append(dict(face_idx=output_ids[i], face_label='', face_name=face_name, face_scores=output_scores, face_image_type=self.image_type, face_image_shape=face_image_npy.shape, face_embedding=output_embeddings[i], face_image=img_b64))
        return result, output_image
    


from iinfer.app import common, predict, train, injection
from pathlib import Path
from typing import List, Dict, Any, Tuple
import importlib.util
import inspect
import iinfer
import logging
import pkgutil
import iinfer


def load_custom_predict(custom_predict_py:Path, logger:logging.Logger) -> predict.Predict:
    """
    カスタム推論オブジェクトを読み込みます。

    Args:
        custom_predict_py (Path): カスタム推論オブジェクトのパス
        logger (logging.Logger): ロガー

    Raises:
        BaseException: 指定されたオブジェクトが見つからない場合

    Returns:
        iinfer.app.predict.Predict: 推論オブジェクト
    """
    spec = importlib.util.spec_from_file_location("iinfer.user.predict", custom_predict_py)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    for name, obj in inspect.getmembers(module):
        if hasattr(obj, "__module__") and obj.__module__ == module.__name__ and inspect.isclass(obj) and issubclass(obj, predict.Predict):
            width = getattr(module, 'IMAGE_WIDTH')
            height = getattr(module, 'IMAGE_HEIGHT')
            pred = obj(logger)
            pred.IMAGE_WIDTH = width
            pred.IMAGE_HEIGHT = height
            return pred
    raise BaseException(f"Predict class not found.({custom_predict_py})")

def load_custom_train(custom_train_py:Path, logger:logging.Logger) -> train.Train:
    """
    カスタム学習オブジェクトを読み込みます。

    Args:
        custom_train_py (Path): カスタム学習オブジェクトのパス
        logger (logging.Logger): ロガー

    Raises:
        BaseException: 指定されたオブジェクトが見つからない場合

    Returns:
        iinfer.app.train.Train: 学習オブジェクト
    """
    spec = importlib.util.spec_from_file_location("iinfer.user.train", custom_train_py)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    for name, obj in inspect.getmembers(module):
        if hasattr(obj, "__module__") and obj.__module__ == module.__name__ and inspect.isclass(obj) and issubclass(obj, train.Train):
            return obj(logger)
    raise BaseException(f"Train class not found.({custom_train_py})")

def build_predict(predict_type:str, custom_predict_py:str, logger:logging.Logger) -> Tuple[bool, predict.Predict]:
    """
    推論オブジェクトを構築します。

    Args:
        predict_type (str): 推論オブジェクトの型
        custom_predict_py (str): カスタム推論オブジェクトのパス
        logger (logging.Logger): ロガー
    
    Returns:
        Tuple[bool, predict.Predict]: 成功した場合はTrueと推論オブジェクト、失敗した場合はFalseとエラーメッセージ
    """
    if predict_type == 'Custom':
        custom_predict_py = Path(custom_predict_py) if custom_predict_py is not None else None
        if custom_predict_py is None:
            logger.warn(f"predict_type is Custom but custom_predict_py is None.")
            return False, {"warn": f"predict_type is Custom but custom_predict_py is None."}
        if not custom_predict_py.exists():
            logger.warn(f"custom_predict_py path {str(custom_predict_py)} does not exist")
            return False, {"warn": f"custom_predict_py path {str(custom_predict_py)} does not exist"}
        predict_obj = load_custom_predict(custom_predict_py, logger)
    else:
        predict_obj = load_predict(predict_type, logger)
    return True, predict_obj

def build_train(train_type:str, custom_train_py:str, logger:logging.Logger) -> Tuple[bool, train.Train]:
    """
    学習オブジェクトを構築します。

    Args:
        train_type (str): 学習オブジェクトの型
        custom_train_py (str): カスタム学習オブジェクトのパス
        logger (logging.Logger): ロガー
    
    Returns:
        Tuple[bool, train.Train]: 成功した場合はTrueと学習オブジェクト、失敗した場合はFalseとエラーメッセージ
    """
    if train_type == 'Custom':
        custom_train_py = Path(custom_train_py) if custom_train_py is not None else None
        if custom_train_py is None:
            logger.warn(f"train_type is Custom but custom_train_py is None.")
            return False, {"warn": f"train_type is Custom but custom_train_py is None."}
        if not custom_train_py.exists():
            logger.warn(f"custom_train_py path {str(custom_train_py)} does not exist")
            return False, {"warn": f"custom_train_py path {str(custom_train_py)} does not exist"}
        train_obj = load_custom_train(custom_train_py, logger)
    else:
        train_obj = load_train(train_type, logger)
    return True, train_obj

def load_predict(predict_type:str, logger:logging.Logger) -> predict.Predict:
    """
    指定された推論オブジェクトを読み込みます。

    Args:
        predict_type (str): 推論オブジェクトのモジュール名
        logger (logging.Logger): ロガー

    Raises:
        BaseException: 指定されたオブジェクトが見つからない場合

    Returns:
        predict.Predict: 推論オブジェクト
    """
    module = importlib.import_module("iinfer.app.predicts." + predict_type)
    for name, obj in inspect.getmembers(module):
        if hasattr(obj, "__module__") and obj.__module__ == module.__name__ and inspect.isclass(obj) and issubclass(obj, predict.Predict):
            return obj(logger)
    raise BaseException(f"Predict class not found.({predict_type})")

def load_train(train_type:str, logger:logging.Logger) -> train.Train:
    """
    指定された学習オブジェクトを読み込みます。

    Args:
        train_type (str): 学習オブジェクトのモジュール名
        logger (logging.Logger): ロガー

    Raises:
        BaseException: 指定されたオブジェクトが見つからない場合

    Returns:
        train.Train: 学習オブジェクト
    """
    module = importlib.import_module("iinfer.app.trains." + train_type)
    for name, obj in inspect.getmembers(module):
        if hasattr(obj, "__module__") and obj.__module__ == module.__name__ and inspect.isclass(obj) and issubclass(obj, train.Train):
            return obj(logger)
    raise BaseException(f"Train class not found.({train_type})")

def load_before_injection_type(injection_type:List[str], config:Dict[str,Any], logger:logging.Logger) -> List[injection.BeforeInjection]:
    """
    指定された前処理オブジェクトを読み込みます。

    Args:
        injection_type (List[str]): 前処理オブジェクトのモジュール名
        config (Dict[str,Any]): 設定
        logger (logging.Logger): ロガー

    Raises:
        BaseException: 指定されたオブジェクトが見つからない場合

    Returns:
        List[injection.BeforeInjection]: 前処理オブジェクト
    """
    injections = []
    for t in injection_type:
        module = importlib.import_module("iinfer.app.injections." + t)
        for name, obj in inspect.getmembers(module):
            if hasattr(obj, "__module__") and obj.__module__ == module.__name__ and inspect.isclass(obj) and issubclass(obj, injection.BeforeInjection):
                injections.append(obj(config, logger))
    if len(injections) > 0:
        return injections
    raise BaseException(f"BeforeInjection class not found.({injection_type})")

def load_after_injection_type(injection_type:List[str], config:Dict[str,Any], logger:logging.Logger) -> List[injection.AfterInjection]:
    """
    指定された後処理オブジェクトを読み込みます。

    Args:
        injection_type (List[str]): 後処理オブジェクトのモジュール名
        config (Dict[str,Any]): 設定
        logger (logging.Logger): ロガー

    Raises:
        BaseException: 指定されたオブジェクトが見つからない場合

    Returns:
        List[injection.AfterInjection]: 後処理オブジェクト
    """
    injections = []
    for t in injection_type:
        module = importlib.import_module("iinfer.app.injections." + t)
        for name, obj in inspect.getmembers(module):
            if hasattr(obj, "__module__") and obj.__module__ == module.__name__ and inspect.isclass(obj) and issubclass(obj, injection.AfterInjection):
                injections.append(obj(config, logger))
    if len(injections) > 0:
        return injections
    raise BaseException(f"AfterInjection class not found.({injection_type})")

def load_before_injections(before_injection_py:List[Path], config:Dict[str,Any], logger:logging.Logger) -> List[injection.BeforeInjection]:
    """
    前処理オブジェクトを読み込みます。

    Args:
        before_injection_py (List[Path]): 前処理オブジェクトのパス
        config (Dict[str,Any]): 設定
        logger (logging.Logger): ロガー

    Returns:
        List[injection.BeforeInjection]: 前処理オブジェクト
    """
    injections = []
    for p in before_injection_py:
        spec = importlib.util.spec_from_file_location("iinfer.user.injection", p)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        for name, obj in inspect.getmembers(module):
            if hasattr(obj, "__module__") and obj.__module__ == module.__name__ and inspect.isclass(obj) and issubclass(obj, injection.BeforeInjection):
                injections.append(obj(config, logger))
    return injections


def load_after_injections(after_injection_py:List[Path], config:Dict[str,Any], logger:logging.Logger) -> List[injection.AfterInjection]:
    """
    後処理オブジェクトを読み込みます。

    Args:
        after_injection_py (List[Path]): 後処理オブジェクトのパス
        config (Dict[str,Any]): 設定
        logger (logging.Logger): ロガー

    Returns:
        List[injection.AfterInjection]: 後処理オブジェクト
    """
    injections = []
    for p in after_injection_py:
        spec = importlib.util.spec_from_file_location("iinfer.user.injection", p)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        for name, obj in inspect.getmembers(module):
            if hasattr(obj, "__module__") and obj.__module__ == module.__name__ and inspect.isclass(obj) and issubclass(obj, injection.AfterInjection):
                injections.append(obj(config, logger))
    return injections

def get_module_list(package_name) -> List[str]:
    """
    パッケージ内のモジュール名のリストを取得します。

    Args:
        package_name (str): パッケージ名

    Returns:
        List[str]: モジュール名のリスト
    """
    package = __import__(package_name, fromlist=[''])
    return [name for _, name, _ in pkgutil.iter_modules(package.__path__)]

for mod in get_module_list('iinfer.app.predicts'):
    if mod.startswith('__'):
        continue
    m = importlib.import_module("iinfer.app.predicts." + mod)
    site = None
    width = None
    height = None
    model_type = iinfer.app.predict.Predict
    required_model_conf = False
    required_model_weight = False
    for f in dir(m):
        members = inspect.getmembers(m, inspect.isclass)
        for name, cls in members:
            if issubclass(cls, iinfer.app.predict.Predict):
                model_type = cls
                break
        if f == 'SITE': site = getattr(m, f)
        elif f == 'IMAGE_WIDTH': width = getattr(m, f)
        elif f == 'IMAGE_HEIGHT': height = getattr(m, f)
        elif f == 'MODEL_TYPE': model_type = getattr(m, f)
        elif f == 'REQUIREd_MODEL_CONF': required_model_conf = getattr(m, f)
        elif f == 'REQUIREd_MODEL_WEIGHT': required_model_weight = getattr(m, f)
    common.BASE_MODELS[mod] = dict(site=site, image_width=width, image_height=height, model_type=model_type,
                                   required_model_conf=required_model_conf, required_model_weight=required_model_weight)

for mod in get_module_list('iinfer.app.trains'):
    if mod.startswith('__'):
        continue
    m = importlib.import_module("iinfer.app.trains." + mod)
    site = None
    model_type = iinfer.app.train.Train
    for f in dir(m):
        members = inspect.getmembers(m, inspect.isclass)
        for name, cls in members:
            if issubclass(cls, iinfer.app.train.Train):
                model_type = cls
                break
        if f == 'SITE': site = getattr(m, f)
    common.BASE_TRAIN_MODELS[mod] = dict(site=site, model_type=model_type)

for mod in get_module_list('iinfer.app.injections'):
    if mod.startswith('__'):
        continue
    try:
        load_before_injection_type([mod], dict(), None)
        common.BASE_BREFORE_INJECTIONS[mod] = dict()
    except:
        pass
    try:
        load_after_injection_type([mod], dict(), None)
        common.BASE_AFTER_INJECTIONS[mod] = dict()
    except:
        pass



from iinfer.app import predict
from iinfer.app.commons import convert
from pathlib import Path
from PIL import Image
from typing import List, Tuple, Union, Any
import logging
import numpy as np


SITE = 'https://github.com/onnx/models/tree/main/vision/classification/efficientnet-lite4'
IMAGE_WIDTH = 224
IMAGE_HEIGHT = 224
REQUIREd_MODEL_CONF = False
REQUIREd_MODEL_WEIGHT = True

class OnnxClsEfficientNetLite4(predict.OnnxPredict):
    def __init__(self, logger:logging.Logger) -> None:
        super().__init__(logger)

    def create_session(self, deploy_dir:Path, model_path:Union[Path,Any], model_conf_path:Path, model_provider:str, gpu_id:int=None):
        """
        推論セッションを作成する関数です。
        startコマンド実行時に呼び出されます。
        この関数内でAIモデルのロードを行い、推論準備を完了するようにしてください。
        戻り値の推論セッションの型は問いません。

        Args:
            deploy_dir (Path): デプロイディレクトリのパス
            model_path (Path|Any): モデルファイルのパス
            model_conf_path (Path): モデル設定ファイルのパス
            gpu_id (int, optional): GPU ID. Defaults to None.

        Returns:
            推論セッション
        """
        import onnxruntime as rt
        if self.is_gpu_available(model_path, model_conf_path, gpu_id):
            session = rt.InferenceSession(model_path, providers=[model_provider])
        else:
            session = rt.InferenceSession(model_path, providers=[model_provider], providers_options=[{'device_id': str(gpu_id)}])
        return session

    def predict(self, model, img_width:int, img_height:int, input_data:Union[Image.Image, str], labels:List[str]=None, colors:List[Tuple[int]]=None, nodraw:bool=False):
        """
        予測を行う関数です。
        predictコマンドやcaptureコマンド実行時に呼び出されます。
        引数のinput_dataが画像の場合RGBですので、戻り値の出力画像もRGBにしてください。
        戻り値の推論結果のdictは、通常推論結果項目ごとに値(list)を設定します。

        Args:
            model: 推論セッション
            img_width (int): モデルのINPUTサイズ（input_dataが画像の場合は、画像の幅）
            img_height (int): モデルのINPUTサイズ（input_dataが画像の場合は、画像の高さ）
            input_data (Image | str): 推論するデータ（画像の場合RGB配列であること）
            labels (List[str], optional): クラスラベルのリスト. Defaults to None.
            colors (List[Tuple[int]], optional): ボックスの色のリスト. Defaults to None.
            nodraw (bool, optional): 描画フラグ. Defaults to False.

        Returns:
            Tuple[Dict[str, Any], Image]: 予測結果と出力画像(RGB)のタプル
        """
        # RGB画像をBGR画像に変換
        img_npy = convert.img2npy(input_data)
        img_npy = convert.bgr2rgb(img_npy)

        image_data, _, image_obj = self.preprocess_img(input_data, img_width, img_height)

        results = model.run(["Softmax:0"], {"images:0": image_data})[0]

        output_scores, output_classes = [], []
        result = reversed(results[0].argsort()[-5:])
        for r in result:
            output_classes.append(r)
            output_scores.append(results[0][r])

        return dict(output_scores=output_scores, output_classes=output_classes), image_obj

    def resize_img(self, image:Image.Image, to_w, to_h):
        '''resize image with unchanged aspect ratio using padding'''
        iw, ih = image.size
        scale = min(to_w/iw, to_h/ih)
        nw = int(iw*scale)
        nh = int(ih*scale)
        image = image.resize((nw,nh), Image.BICUBIC)
        new_image = Image.new('RGB', (to_w, to_h), (128,128,128))
        new_image.paste(image, ((to_w-nw)//2, (to_h-nh)//2))
        return new_image

    def preprocess_img(self, image:Image.Image, model_img_width:int, model_img_height:int):
        boxed_image = self.resize_img(image, model_img_width, model_img_height)
        image_data = np.array(boxed_image, dtype='float32')
        image_data /= 255.
        image_data = np.expand_dims(image_data, 0)
        image_size = np.array([image.size[1], image.size[0]], dtype=np.float32).reshape(1, 2)
        return image_data, image_size, image



from iinfer.app import postprocess
from iinfer.app.commons import convert
from iinfer.app.injections import after_det_jadge_injection
from PIL import Image
from typing import Dict, Any, Tuple, List
import cv2
import logging

class DetJadge(postprocess.Postprocess):
    def __init__(self, logger:logging.Logger,
                 ok_score_th:float=None, ok_classes:List[int]=None, ok_labels:List[str]=None,
                 ng_score_th:float=None, ng_classes:List[int]=None, ng_labels:List[str]=None,
                 ext_score_th:float=None, ext_classes:List[int]=None, ext_labels:List[str]=None,
                 nodraw:bool=False, output_preview:bool=False):
        """
        Object Detection推論結果のクラススコア元に、この画像としてOK/NG/Grayの判定結果を追加する後処理クラスです。

        Args:
            logger (logging.Logger): ロガー
            ok_score_th (float): ok判定確定のスコアの閾値
            ok_classes (List[int]): ok判定確定のクラスのリスト
            ok_labels (List[str]): ok判定確定のラベルのリスト
            ng_score_th (float): ng判定確定のスコアの閾値
            ng_classes (List[int]): ng判定確定のクラスのリスト
            ng_labels (List[str]): ng判定確定のラベルのリスト
            ext_score_th (float): gray判定確定のスコアの閾値
            ext_classes (List[int]): gray判定確定のクラスのリスト
            ext_labels (List[str]): gray判定確定のラベルのリスト
            nodraw (bool): 描画しない
            output_preview (bool): プレビューを出力する
        """
        super().__init__(logger)
        self.config = dict(ok_score_th=ok_score_th, ok_classes=ok_classes, ok_labels=ok_labels,
                           ng_score_th=ng_score_th, ng_classes=ng_classes, ng_labels=ng_labels,
                           ext_score_th=ext_score_th, ext_classes=ext_classes, ext_labels=ext_labels,
                           nodraw=nodraw, output_preview=output_preview)
        self.injection = after_det_jadge_injection.AfterDetJadgeInjection(self.config, self.logger)

    def post(self, outputs:Dict[str, Any], output_image:Image.Image) -> Tuple[Dict[str, Any], Image.Image]:
        """
        後処理を行う関数です。
        outputsは、以下のような構造を持つDict[str, Any]です。
        {
            'success': {
                'output_ids': List[int],
                'output_scores': List[float],
                'output_classes': List[int],
                'output_labels': List[str],
                'output_boxes': List[List[int]],
                'output_tracks': List[int]
            }
        }

        Args:
            outputs (Dict[str, Any]): 推論結果
            output_image (Image.Image): 入力画像（RGB配列であること）

        Returns:
            Dict[str, Any]: 後処理結果
            Image: 後処理結果
        """
        outputs, output_image = self.injection.action(None, None, outputs, output_image, None)
        output_preview = self.injection.get_config('output_preview', False)
        if output_preview:
            # RGB画像をBGR画像に変換
            img_npy = convert.img2npy(output_image).copy()
            img_npy = convert.bgr2rgb(img_npy)
            try:
                cv2.imshow('preview', img_npy)
                cv2.waitKey(1)
            except KeyboardInterrupt:
                pass
        return outputs, output_image



from pathlib import Path
from PIL import Image
from typing import List, Tuple, Dict, Any, Union
import logging


class Predict(object):
    def __init__(self, logger:logging.Logger) -> None:
        """
        このクラスのインスタンスを初期化します。
        継承時は、このコンストラクタを呼び出すようにしてください。
            super().__init__(logger)
        Args:
            logger (logging.Logger): ロガー
        """
        self.logger = logger

    def is_gpu_available(self, gpu_id:int=None) -> bool:
        """
        GPUが利用可能かどうかを返す関数です。
        戻り値がTrueの場合、GPUが利用可能です。
        戻り値がFalseの場合、GPUが利用不可です。

        Args:
            gpu_id (int, optional): GPU ID. Defaults to None.
        Returns:
            bool: GPUが利用可能かどうか
        """
        raise NotImplementedError()

    def post_deploy(self, deploy_dir:Path, conf:dict) -> None:
        """
        デプロイ後の処理を行う関数です。
        deployコマンド実行時に呼び出されます。
        この関数内でデプロイ後の処理を実装してください。
        
        Args:
            deploy_dir (Path): デプロイディレクトリのパス
            conf (dict): デプロイ設定
        """
        pass

    def create_session(self, deploy_dir:Path, model_path:Union[Path,Any], model_conf_path:Path, model_provider:str, gpu_id:int=None) -> Any:
        """
        推論セッションを作成する関数です。
        startコマンド実行時に呼び出されます。
        この関数内でAIモデルのロードを行い、推論準備を完了するようにしてください。
        戻り値の推論セッションの型は問いません。

        Args:
            deploy_dir (Path): デプロイディレクトリのパス
            model_path (Path|Any): モデルファイルのパス
            model_conf_path (Path): モデル設定ファイルのパス
            gpu_id (int, optional): GPU ID. Defaults to None.

        Returns:
            推論セッション
        """
        raise NotImplementedError()

    def predict(self, session, img_width:int, img_height:int, image:Image.Image, labels:List[str]=None, colors:List[Tuple[int]]=None, nodraw:bool=False) -> Tuple[Dict[str, Any], Image.Image]:
        """
        予測を行う関数です。
        predictコマンドやcaptureコマンド実行時に呼び出されます。
        引数のimageはRGBですので、戻り値の出力画像もRGBにしてください。
        戻り値の推論結果のdictは、通常推論結果項目ごとに値(list)を設定します。
        例）Image Classification（EfficientNet_Lite4）の場合
        return dict(output_scores=output_scores, output_classes=output_classes), image_obj
        例）Object Detection（YoloX）の場合
        return dict(output_boxes=final_boxes, output_scores=final_scores, output_classes=final_cls_inds), output_image

        Args:
            session: 推論セッション
            img_width (int): モデルのINPUTサイズ（画像の幅）
            img_height (int): モデルのINPUTサイズ（画像の高さ）
            image (Image): 入力画像（RGB配列であること）
            labels (List[str], optional): クラスラベルのリスト. Defaults to None.
            colors (List[Tuple[int]], optional): ボックスの色のリスト. Defaults to None.
            nodraw (bool, optional): 描画フラグ. Defaults to False.

        Returns:
            Tuple[Dict[str, Any], Image]: 予測結果と出力画像(RGB)のタプル
        """
        raise NotImplementedError()

class OnnxPredict(Predict):
    def __init__(self, logger:logging.Logger) -> None:
        """
        このクラスのインスタンスを初期化します。
        継承時は、このコンストラクタを呼び出すようにしてください。
            super().__init__(logger)
        Args:
            logger (logging.Logger): ロガー
        """
        super().__init__(logger)

    def is_gpu_available(self, model_path:Union[Path,Any], model_conf_path:Path, gpu_id:int=None) -> bool:
        """
        GPUが利用可能かどうかを返す関数です。
        戻り値がTrueの場合、GPUが利用可能です。
        戻り値がFalseの場合、GPUが利用不可です。

        Args:
            model_path (Path|Any): モデルファイルのパス
            model_conf_path (Path): モデル設定ファイルのパス
            gpu_id (int, optional): GPU ID. Defaults to None.
        Returns:
            bool: GPUが利用可能かどうか
        """
        try:
            import onnxruntime as rt
            rt.InferenceSession(model_path, providers=["CUDAExecutionProvider"], providers_options=[{'device_id': str(gpu_id)}])
            self.logger.info(f'GPU is available: True')
            return True
        except:
            self.logger.info(f'GPU is available: False')
            return False

class TorchPredict(Predict):
    def __init__(self, logger:logging.Logger) -> None:
        """
        このクラスのインスタンスを初期化します。
        継承時は、このコンストラクタを呼び出すようにしてください。
            super().__init__(logger)
        Args:
            logger (logging.Logger): ロガー
        """
        super().__init__(logger)

    def is_gpu_available(self, model_path:Union[Path,Any], model_conf_path:Path, gpu_id:int=None) -> bool:
        """
        GPUが利用可能かどうかを返す関数です。
        戻り値がTrueの場合、GPUが利用可能です。
        戻り値がFalseの場合、GPUが利用不可です。

        Args:
            model_path (Path|Any): モデルファイルのパス
            model_conf_path (Path): モデル設定ファイルのパス
            gpu_id (int, optional): GPU ID. Defaults to None.
        Returns:
            bool: GPUが利用可能かどうか
        """
        try:
            import torch
            ret = torch.cuda.is_available()
            self.logger.info(f'GPU is available: {ret}')
            return ret
        except:
            self.logger.info(f'GPU is available: False')
            return False




from iinfer.app import options, injection
from iinfer.app.commons import convert, redis_client
from PIL import Image
from typing import Tuple, Dict, Any
import logging

class AfterShowimgInjection(injection.AfterInjection):
    def __init__(self, config:Dict[str,Any], logger:logging.Logger):
        super().__init__(config, logger)
        opts = options.Options()
        self.default_host = opts.get_cmd_opt('server', 'start', 'host')['default']
        self.default_port = opts.get_cmd_opt('server', 'start', 'port')['default']
        self.default_password = opts.get_cmd_opt('server', 'start', 'password')['default']
        self.default_svname = opts.get_cmd_opt('server', 'start', 'svname')['default']


    def action(self, reskey:str, name:str, outputs:Dict[str, Any], output_image:Image.Image, session:Dict[str, Any]) -> Tuple[Dict[str, Any], Image.Image]:
        """
        このメソッドは推論を実行した後の処理を実行します。
        Args:
            reskey (str): レスポンスキー
            name (str): モデル名
            outputs (Dict[str, Any]): 推論結果。次の項目が含まれます。
                ・success or warn: 推論成功か警告のキーに対して、その内容が格納されます。
                ・output_image: 推論後の画像データをbase64エンコードした文字列
                ・output_image_shape: 推論後の画像データの形状（base46でコードするときに必要）
                ・output_image_name: クライアントから指定されてきた推論後の画像データの名前

            output_image (Image.Image): 推論後の画像データ
            session (Dict[str, Any]): 推論セッション。次の項目が含まれます。
                ・session: app.predict.Predict#create_session() で作成されたセッション
                ・model_img_width: モデルの入力画像の幅
                ・model_img_height: モデルの入力画像の高さ
                ・predict_obj: app.predict.Predict インスタンス
                ・labels: クラスラベルのリスト
                ・colors: ボックスの色のリスト
                ・tracker: use_trackがTrueの場合、トラッカーのインスタンス
        Returns:
            Tuple[Dict[str, Any], Image.Image]: 後処理後の推論結果と画像データのタプル
        """
        host = self.get_config('host', self.default_host)
        port = self.get_config('port', self.default_port)
        password = self.get_config('password', self.default_password)
        svname = self.get_config('svname', self.default_svname)
        maxrecsize = self.get_config('maxrecsize', 200)
        self.redis_cli = redis_client.RedisClient(self.logger, host=host, port=port, password=password, svname=svname)

        try:
            img_npy = convert.img2npy(output_image)
            outputs['output_image'] = convert.npy2b64str(img_npy)
            outputs['output_image_shape'] = img_npy.shape
            result = self.redis_cli.send_showimg('outputs', outputs, maxrecsize=maxrecsize)
            self.add_success(outputs, result)
            del outputs['output_image']
            del outputs['output_image_shape']
        except Exception as e:
            self.add_warning(outputs, str(e))

        return outputs, output_image



from iinfer.app.trains import mmseg_seg_PSPNet
import logging


SITE = 'https://github.com/open-mmlab/mmsegmentation/tree/main/configs/swin'

class MMSegSwinUpernet(mmseg_seg_PSPNet.MMSegPSPNet):
    def __init__(self, logger:logging.Logger) -> None:
        super().__init__(logger)



from iinfer.app import common
from iinfer.app.commons import convert
from PIL import Image
from typing import List, Dict, Any
import datetime
import logging
import json
import redis
import time


class RedisClient(object):
    def __init__(self, logger:logging.Logger, host:str = "localhost", port:int = 6379, password:str = None, svname:str="server"):
        """
        Redisクライアントを初期化します。
        
        Args:
            logger (Logger): ロガー
            host (str): Redisのホスト名
            port (int): Redisのポート番号
            password (str): Redisのパスワード
            svname (str): サーバー名
        """
        self.logger = logger
        self.host = host
        self.port = port
        self.password = password
        self.svname = f"sv-{svname}"
        self.hbname = f"hb-{svname}"
        self.siname = f"showimg-{svname}"
        self.redis_cli = self.connect()

    def connect(self):
        """
        Redisに接続します。
        """
        return redis.Redis(host=self.host, port=self.port, db=0, password=self.password)
    
    def close(self):
        if self.redis_cli is not None:
            self.redis_cli.close()
            self.redis_cli = None
    
    def rpush(self, name:str, value:dict):
        if type(value) is dict or type(value) is list:
            self.redis_cli.rpush(name, json.dumps(value, default=common.default_json_enc))
        elif type(value) is str:
            self.redis_cli.rpush(name, value)
        else:
            self.logger.warning(f"Unsupported type. {type(value)}")
            raise Exception(f"Unsupported type. {type(value)}")

    def blpop(self, name:str, timeout:int=1):
        return self.redis_cli.blpop(name, timeout=timeout)

    def lpop(self, name:str):
        return self.redis_cli.lpop(name)

    def hset(self, name:str, key:str, value):
        self.redis_cli.hset(name, key, str(value))

    def delete(self, name:str):
        self.redis_cli.delete(name)
    
    def hexists(self, name:str, key:str):
        return self.redis_cli.hexists(name, key)
    
    def hget(self, name:str, key:str):
        return self.redis_cli.hget(name, key)

    def keys(self, pattern:str):
        return self.redis_cli.keys(pattern)

    def check_server(self, find_svname:bool=False, retry_count:int=20, retry_interval:int=5, outstatus:bool=False):
        """
        Redisサーバーにpingを送信し、応答があるか確認する

        Args:
            find_svname (bool, optional): サーバー名が存在するか確認する. Defaults to False.
            retry_count (int, optional): リトライ回数. Defaults to 20.
            retry_interval (int, optional): リトライ間隔. Defaults to 5.
            outstatus (bool, optional): ステータスを出力する. Defaults to False.

        Returns:
            bool: 接続成功時はTrue、失敗時はFalse
        """
        if retry_interval <= 0:
            self.logger.warning(f"retry_interval must be greater than 0. retry_interval={retry_interval}", exc_info=True)
            return False

        i = 0
        while i < retry_count or retry_count <= 0:
            try:
                if outstatus:
                    self.logger.info(f"({i+1}/{retry_count if retry_count>0 else '-'}) connecting to the redis server. {self.host}:{self.port}")
                self.redis_cli.ping()
                if find_svname:
                    _hbname = f"{self.hbname}-*" if len(self.hbname.split('-')) < 3 else self.hbname
                    found = self.redis_cli.keys(_hbname)
                    if len(found) <= 0:
                        self.logger.warning(f"Server not found. svname={self.svname.split('-')[1]}")
                        return False
                i = 0
                return True
            except redis.exceptions.ConnectionError as e:
                self.logger.warning(f"fail to ping server. {e}")
                if i >= retry_count and retry_count > 0:
                    return False
                time.sleep(retry_interval if retry_interval > 0 else 5)
                i += 1
            except KeyboardInterrupt as e:
                return False

    def send_cmd(self, cmd:str, params:List[str], retry_count:int=20, retry_interval:int=5, outstatus:bool=False, timeout:int = 60):
        """
        コマンドをRedisサーバーに送信し、応答を取得する

        Args:
            cmd (str): コマンド
            params (List[str]): コマンドのパラメータ
            retry_count (int, optional): リトライ回数. Defaults to 20.
            retry_interval (int, optional): リトライ間隔. Defaults to 5.
            outstatus (bool, optional): ステータスを出力する. Defaults to False.
            timeout (int, optional): タイムアウト時間. Defaults to 60.

        Returns:
            dict: Redisサーバーからの応答
        """
        try:
            if timeout <= 0:
                raise ValueError(f"timeout must be greater than 0. timeout={timeout}")
            sreqtime = time.perf_counter()
            if not self.check_server(find_svname=True, retry_count=retry_count, retry_interval=retry_interval, outstatus=outstatus):
                return dict(error=f"Connected server failed or server not found. svname={self.svname.split('-')[1]}")
            reskey = common.random_string()
            reskey = f"cl-{reskey}-{int(time.time())}"
            self.redis_cli.rpush(self.svname, f"{cmd} {reskey} {' '.join([str(p) for p in params])}")
            self.is_running = True
            stime = time.time()
            while self.is_running:
                ctime = time.time()
                if ctime - stime > timeout:
                    raise Exception(f"Response timed out.")
                res = self.redis_cli.lpop(reskey)
                if res is None or len(res) <= 0:
                    time.sleep(0.0001)
                    continue
                return self._res_cmd(reskey, res, sreqtime)
            raise KeyboardInterrupt(f"Stop command.")
        except KeyboardInterrupt as e:
            self.logger.warning(f"Stop command. cmd={cmd}", exc_info=True)
            return dict(error=f"Stop command. cmd={cmd}")
        except Exception as e:
            self.logger.warning(f"fail to execute command. cmd={cmd}, msg={e}", exc_info=True)
            return dict(error=f"fail to execute command. cmd={cmd}, msg={e}")

    def _res_cmd(self, reskey:str, res_msg:bytes, sreqtime:float):
        """
        Redisサーバーからの応答を解析する

        Args:
            reskey (str): Redisサーバーからの応答のキー
            res_msg (bytes): Redisサーバーからの応答

        Returns:
            dict: 解析された応答
        """
        self.redis_cli.delete(reskey)
        reskbyte = len(res_msg) / 1024
        msg = res_msg.decode('utf-8')
        res_json = json.loads(msg)
        msg_json = res_json.copy()
        if "output_image" in msg_json:
            msg_json["output_image"] = "binary"
        if "error" in res_json:
            self.logger.warning(common.to_str(msg_json))
        if "warn" in res_json:
            self.logger.warning(common.to_str(msg_json))
        if "success" in res_json:
            if self.logger.level == logging.DEBUG:
                msg_str = common.to_str(msg_json, slise=100)
                self.logger.debug(f"redis_client._res_cmd: msg={msg_str}")
            if type(res_json["success"]) is not dict:
                res_json["success"] = dict(data=res_json["success"])
            if "performance" not in res_json["success"]:
                res_json["success"]["performance"] = []
            performance = res_json["success"]["performance"]
            performance.append(dict(key="cl_svreqest", val=f"{time.perf_counter()-sreqtime:.3f}s"))
            performance.append(dict(key="cl_reskbyte", val=f"{reskbyte:.3f}KB"))
        return res_json

    def receive_showimg(self):
        if not self.check_server(find_svname=False):
            return None, None
        result = self.lpop(self.siname)
        if result is None or len(result) <= 0:
            return None, None
        msg = result.decode().split(' ')
        if self.logger.level == logging.DEBUG:
            msg_str = common.to_str(msg, slise=100)
            self.logger.debug(f"redis_client.receive_showimg: self.siname={self.siname}, msg={msg_str}")
        if len(msg) <= 0:
            return None, None
        cmd = msg[0]
        if cmd == "text":
            value = convert.b64str2str(msg[1])
            return cmd, value
        elif cmd == "outputs":
            value = convert.b64str2str(msg[1])
            ret = cmd, json.loads(value)
            return ret
        elif cmd == "output_image":
            line = convert.b64str2str(msg[1])
            capture_data = line.split(',')
            t = capture_data[0]
            img = capture_data[1]
            h = int(capture_data[2])
            w = int(capture_data[3])
            c = int(capture_data[4])
            filename = capture_data[5]
            img_npy = convert.b64str2npy(img, shape=(h, w, c) if c > 0 else (h, w))
            img_bytes = convert.npy2imgfile(img_npy)
            return cmd, (filename, img_bytes)
        else:
            self.logger.warning(f"Unsupported cmd. cmd={cmd}")
            return None, None

    def send_showimg(self, cmd:str, outputs:Dict[str, Any], output_image:Image.Image=None, maxrecsize:int=200):
        """
        後処理結果をRedisサーバーに送信する

        Args:
            cmd (str): コマンド
            outputs (Dict[str, Any]): 後処理結果
            output_image (Image.Image, optional): 画像データ. Defaults to None.
            maxrecsize (int, optional): 最大レコードサイズ. Defaults to 200.
        """
        if cmd not in ["text", "outputs", "output_image"]:
            self.logger.warning(f"Unsupported cmd. cmd={cmd}")
            return dict(warn=f"Unsupported cmd. cmd={cmd}")
        if not self.check_server(find_svname=True):
            return dict(error=f"Server chack failed. svname={self.svname.split('-')[1]}")
        llen = self.redis_cli.llen(self.siname)
        if llen > maxrecsize:
            self.logger.warning(f"Cancelled execution of showimg. '{self.siname}' list is full. llen={llen} maxrecsize={maxrecsize}")
            return dict(warn=f"Cancelled execution of showimg. '{self.siname}' list is full. llen={llen} maxrecsize={maxrecsize}")
        try:
            if cmd == "outputs":
                if self.logger.level == logging.DEBUG:
                    outputs_str = common.to_str(outputs, slise=100)
                    self.logger.debug(f"redis_client.send_showimg: self.siname={self.siname}, cmd={cmd}, outputs={outputs_str}")
                value = convert.str2b64str(json.dumps(outputs, default=common.default_json_enc))
                self.redis_cli.rpush(self.siname, f"{cmd} {value}")
            elif cmd == "output_image":
                img_npy = convert.img2npy(output_image)
                img_b64 = convert.npy2b64str(img_npy)
                output_image_name = outputs['output_image_name'] if 'output_image_name' in outputs else f'{datetime.datetime.now().strftime("%Y%m%d%H%M%S")}.capture'
                img_line = f"capture,{img_b64},{img_npy.shape[0]},{img_npy.shape[1]},{img_npy.shape[2] if len(img_npy.shape) > 2 else -1}, {output_image_name}"
                value = convert.str2b64str(img_line)
                self.redis_cli.rpush(self.siname, f"{cmd} {value}")
            elif cmd == "text":
                value = convert.str2b64str(outputs)
                self.redis_cli.rpush(self.siname, f"{cmd} {value}")
            else:
                self.logger.warning(f"outputs are not supported type. {type(outputs)} cmd='{cmd}' llen={llen}. outputs={outputs}")
                return dict(warn=f"outputs are not supported type. {type(outputs)} cmd='{cmd}' llen={llen}. outputs={outputs}")
        except Exception as e:
            self.logger.warning(f"Failed to publish to Redis server. cmd='{cmd}' llen={llen}. {e}", exc_info=True)
            return dict(warn=f"Failed to publish to Redis server. cmd='{cmd}' llen={llen}. {e}")
        return dict(success=f"Success execution of showimg. cmd='{cmd}' llen={llen+1}.")

    def list_server(self) -> List[Dict[str, Any]]:
        """
        起動しているサーバーリストを取得する

        Returns:
            List[Dict[str, Any]]: サーバーのリスト
        """
        hblist = self.keys("hb-*")
        svlist = []
        for hb in hblist:
            hb = hb.decode()
            svname = hb.replace("hb-", "")
            try:
                val = self.hget(hb, 'receive_cnt')
                receive_cnt = int(val.decode()) if val is not None else 0
                val = self.hget(hb, 'sccess_cnt')
                sccess_cnt = int(val.decode()) if val is not None else 0
                val = self.hget(hb, 'warn_cnt')
                warn_cnt = int(val.decode()) if val is not None else 0
                val = self.hget(hb, 'error_cnt')
                error_cnt = int(val.decode()) if val is not None else 0
                val = self.hget(hb, 'status')
                status = val.decode() if val is not None else "unknown"
                val = self.hget(hb, 'ctime')
                ctime = time.strftime('%Y/%m/%d %H:%M:%S', time.localtime(float(val.decode()))) if val is not None else "-"
            except redis.exceptions.ResponseError:
                self.logger.warn(f"ResponseError. {hb}", exc_info=True)
            svlist.append(dict(svname=svname, status=status, ctime=ctime,
                               receive_cnt=receive_cnt, sccess_cnt=sccess_cnt, warn_cnt=warn_cnt, error_cnt=error_cnt))
        return svlist


from pathlib import Path
from PIL import Image
from iinfer.app import common, train
from iinfer.app.commons import convert
from typing import Dict, Tuple, Union, Any
import argparse
import copy
import logging
import os
import time
import warnings


SITE = 'https://github.com/open-mmlab/mmpretrain/tree/master/configs/swin_transformer'

class MMPretrainClsSwin(train.TorchTrain):
    def __init__(self, logger:logging.Logger) -> None:
        super().__init__(logger)

    def train(self, deploy_dir:Path, model_conf_path:Path, train_cfg_options:Dict[str, Any]=None) -> None:
        """
        学習を実行する関数です。
        trainコマンド実行時に呼び出されます。
        この関数内でAIモデルの学習を行い完了するようにしてください。

        Args:
            deploy_dir (Path): デプロイディレクトリのパス
            model_conf_path (Path): モデル設定ファイルのパス
            train_cfg_options (Dict[str, Any]): 学習設定オプションのリスト
        """
        from mmcv.runner import get_dist_info, init_dist
        from mmcv import Config
        from mmcls import __version__
        from mmcls.apis import init_random_seed, set_random_seed, train_model
        from mmcls.datasets import build_dataset
        from mmcls.models import build_classifier
        from mmcls.utils import (auto_select_device, collect_env, get_root_logger,
                                setup_multi_processes)
        import mmcv
        import torch

        args = argparse.Namespace(**train_cfg_options)

        cfg = Config.fromfile(str(model_conf_path))
        cfg.work_dir = str(deploy_dir / "work_dirs")
        if train_cfg_options is not None:
            cfg.merge_from_dict(train_cfg_options)

        # set multi-process settings
        setup_multi_processes(cfg)

        # set cudnn_benchmark
        if cfg.get('cudnn_benchmark', False):
            torch.backends.cudnn.benchmark = True

        # work_dir is determined in this priority: CLI > segment in file > filename
        if args.work_dir is not None:
            # update configs according to CLI args if args.work_dir is not None
            cfg.work_dir = args.work_dir
        elif cfg.get('work_dir', None) is None:
            # use config filename as default work_dir if cfg.work_dir is None
            cfg.work_dir = os.path.join('./work_dirs',
                                    os.path.splitext(os.path.basename(args.config))[0])
        if args.resume_from is not None:
            cfg.resume_from = args.resume_from
        if args.gpus is not None:
            cfg.gpu_ids = range(1)
            warnings.warn('`--gpus` is deprecated because we only support '
                        'single GPU mode in non-distributed training. '
                        'Use `gpus=1` now.')
        if args.gpu_ids is not None:
            cfg.gpu_ids = args.gpu_ids[0:1]
            warnings.warn('`--gpu-ids` is deprecated, please use `--gpu-id`. '
                        'Because we only support single GPU mode in '
                        'non-distributed training. Use the first GPU '
                        'in `gpu_ids` now.')
        if args.gpus is None and args.gpu_ids is None:
            cfg.gpu_ids = [args.gpu_id]

        if args.ipu_replicas is not None:
            cfg.ipu_replicas = args.ipu_replicas
            args.device = 'ipu'

        # init distributed env first, since logger depends on the dist info.
        if args.launcher == 'none':
            distributed = False
        else:
            distributed = True
            init_dist(args.launcher, **cfg.dist_params)
            _, world_size = get_dist_info()
            cfg.gpu_ids = range(world_size)

        # create work_dir
        mmcv.mkdir_or_exist(os.path.abspath(cfg.work_dir))
        # dump config
        cfg.dump(os.path.join(cfg.work_dir, os.path.basename(args.config)))
        # init the logger before other steps
        timestamp = time.strftime('%Y%m%d_%H%M%S', time.localtime())
        log_file = os.path.join(cfg.work_dir, f'{timestamp}.log')
        logger = get_root_logger(log_file=log_file, log_level=cfg.log_level)

        # init the meta dict to record some important information such as
        # environment info and seed, which will be logged
        meta = dict()
        # log env info
        env_info_dict = collect_env()
        env_info = '\n'.join([(f'{k}: {v}') for k, v in env_info_dict.items()])
        dash_line = '-' * 60 + '\n'
        logger.info('Environment info:\n' + dash_line + env_info + '\n' +
                    dash_line)
        meta['env_info'] = env_info

        # log some basic info
        logger.info(f'Distributed training: {distributed}')
        logger.info(f'Config:\n{cfg.pretty_text}')

        # set random seeds
        cfg.device = args.device or auto_select_device()
        seed = init_random_seed(args.seed, device=cfg.device)
        seed = seed + torch.distributed.get_rank() if args.diff_seed else seed
        logger.info(f'Set random seed to {seed}, '
                    f'deterministic: {args.deterministic}')
        set_random_seed(seed, deterministic=args.deterministic)
        cfg.seed = seed
        meta['seed'] = seed

        model = build_classifier(cfg.model)
        model.init_weights()

        datasets = [build_dataset(cfg.data.train)]
        if len(cfg.workflow) == 2:
            val_dataset = copy.deepcopy(cfg.data.val)
            val_dataset.pipeline = cfg.data.train.pipeline
            datasets.append(build_dataset(val_dataset))

        # save mmcls version, config file content and class names in
        # runner as meta data
        meta.update(
            dict(
                mmcls_version=__version__,
                config=cfg.pretty_text,
                CLASSES=datasets[0].CLASSES))

        # add an attribute for visualization convenience
        train_model(
            model,
            datasets,
            cfg,
            distributed=distributed,
            validate=(not args.no_validate),
            timestamp=timestamp,
            device=cfg.device,
            meta=meta)



from pathlib import Path
from PIL import Image
from iinfer.app import common, predict
from typing import List, Tuple, Union, Any
import logging
import numpy as np


SITE = 'https://github.com/onnx/models/tree/main/vision/object_detection_segmentation/yolov3'
IMAGE_WIDTH = 416
IMAGE_HEIGHT = 416
REQUIREd_MODEL_CONF = False
REQUIREd_MODEL_WEIGHT = True

class OnnxDetYoloV3(predict.OnnxPredict):
    def __init__(self, logger:logging.Logger) -> None:
        super().__init__(logger)

    def create_session(self, deploy_dir:Path, model_path:Union[Path,Any], model_conf_path:Path, model_provider:str, gpu_id:int=None):
        """
        推論セッションを作成する関数です。
        startコマンド実行時に呼び出されます。
        この関数内でAIモデルのロードを行い、推論準備を完了するようにしてください。
        戻り値の推論セッションの型は問いません。

        Args:
            deploy_dir (Path): デプロイディレクトリのパス
            model_path (Path): モデルファイルのパス
            model_conf_path (Path): モデル設定ファイルのパス
            gpu_id (int, optional): GPU ID. Defaults to None.

        Returns:
            推論セッション
        """
        import onnxruntime as rt
        if self.is_gpu_available(model_path, model_conf_path, gpu_id):
            session = rt.InferenceSession(model_path, providers=[model_provider])
        else:
            session = rt.InferenceSession(model_path, providers=[model_provider], providers_options=[{'device_id': str(gpu_id)}])
        return session

    def predict(self, model, img_width:int, img_height:int, input_data:Union[Image.Image, str], labels:List[str]=None, colors:List[Tuple[int]]=None, nodraw:bool=False):
        """
        予測を行う関数です。
        predictコマンドやcaptureコマンド実行時に呼び出されます。
        引数のinput_dataが画像の場合RGBですので、戻り値の出力画像もRGBにしてください。
        戻り値の推論結果のdictは、通常推論結果項目ごとに値(list)を設定します。

        Args:
            model: 推論セッション
            img_width (int): モデルのINPUTサイズ（input_dataが画像の場合は、画像の幅）
            img_height (int): モデルのINPUTサイズ（input_dataが画像の場合は、画像の高さ）
            input_data (Image | str): 推論するデータ（画像の場合RGB配列であること）
            labels (List[str], optional): クラスラベルのリスト. Defaults to None.
            colors (List[Tuple[int]], optional): ボックスの色のリスト. Defaults to None.
            nodraw (bool, optional): 描画フラグ. Defaults to False.

        Returns:
            Tuple[Dict[str, Any], Image]: 予測結果と出力画像(RGB)のタプル
        """
        image_data, image_size, image_obj = self.preprocess_img(input_data, img_width, img_height)

        input_name = model.get_inputs()[0].name           # 'image'
        input_name_img_shape = model.get_inputs()[1].name # 'image_shape'
        output_name_boxes = model.get_outputs()[0].name   # 'boxes'
        output_name_scores = model.get_outputs()[1].name  # 'scores'
        output_name_indices = model.get_outputs()[2].name # 'indices'

        outputs_index = model.run([output_name_boxes, output_name_scores, output_name_indices],
                                  {input_name: image_data, input_name_img_shape: image_size})

        output_boxes = outputs_index[0]
        output_scores = outputs_index[1]
        output_indices = outputs_index[2]

        out_boxes, out_scores, out_classes = [], [], []
        for idx_ in output_indices:
            out_classes.append(idx_[1])
            out_scores.append(output_scores[tuple(idx_)])
            idx_1 = (idx_[0], idx_[2])
            out_boxes.append(output_boxes[idx_1])

        ids = [i for i in range(len(out_boxes))]
        output_image, output_labels = common.draw_boxes(image_obj, out_boxes, out_scores, out_classes, ids=ids, labels=labels, colors=colors, nodraw=nodraw)

        return dict(output_ids=ids, output_scores=out_scores, output_classes=out_classes, output_labels=output_labels, output_boxes=out_boxes), output_image

    def resize_img(self, image:Image.Image, to_w, to_h):
        '''resize image with unchanged aspect ratio using padding'''
        iw, ih = image.size
        scale = min(to_w/iw, to_h/ih)
        nw = int(iw*scale)
        nh = int(ih*scale)
        image = image.resize((nw,nh), Image.BICUBIC)
        new_image = Image.new('RGB', (to_w, to_h), (128,128,128))
        new_image.paste(image, ((to_w-nw)//2, (to_h-nh)//2))
        return new_image

    def preprocess_img(self, image:Image.Image, model_img_width:int, model_img_height:int):
        boxed_image = self.resize_img(image, model_img_width, model_img_height)
        image_data = np.array(boxed_image, dtype='float32')
        image_data /= 255.
        image_data = np.transpose(image_data, [2, 0, 1])
        image_data = np.expand_dims(image_data, 0)
        image_size = np.array([image.size[1], image.size[0]], dtype=np.float32).reshape(1, 2)
        return image_data, image_size, image



from pathlib import Path
from PIL import Image
from iinfer.app import common, predict
from iinfer.app.commons import convert
from typing import List, Tuple, Union, Any
import logging


SITE = 'https://github.com/open-mmlab/mmdetection/tree/main/configs/yolox'
IMAGE_WIDTH = 640
IMAGE_HEIGHT = 640
REQUIREd_MODEL_CONF = True
REQUIREd_MODEL_WEIGHT = False

class MMDetYoloX(predict.TorchPredict):
    def __init__(self, logger:logging.Logger) -> None:
        super().__init__(logger)

    def create_session(self, deploy_dir:Path, model_path:Union[Path,Any], model_conf_path:Path, model_provider:str, gpu_id:int=None):
        """
        推論セッションを作成する関数です。
        startコマンド実行時に呼び出されます。
        この関数内でAIモデルのロードを行い、推論準備を完了するようにしてください。
        戻り値の推論セッションの型は問いません。

        Args:
            deploy_dir (Path): デプロイディレクトリのパス
            model_path (Path|Any): モデルファイルのパス
            model_conf_path (Path): モデル設定ファイルのパス
            gpu_id (int, optional): GPU ID. Defaults to None.

        Returns:
            推論セッション
        """
        from mmdet.apis import init_detector
        import torch
        gpu = f'cuda:{gpu_id}' if gpu_id is not None else 'cuda'
        device = torch.device(gpu if self.is_gpu_available(model_path, model_conf_path, gpu_id) else 'cpu')
        model = init_detector(model_conf_path, str(model_path), device=device) # , cfg_options = {'show': True}
        return model

    def predict(self, model, img_width:int, img_height:int, input_data:Union[Image.Image, str], labels:List[str]=None, colors:List[Tuple[int]]=None, nodraw:bool=False):
        """
        予測を行う関数です。
        predictコマンドやcaptureコマンド実行時に呼び出されます。
        引数のinput_dataが画像の場合RGBですので、戻り値の出力画像もRGBにしてください。
        戻り値の推論結果のdictは、通常推論結果項目ごとに値(list)を設定します。

        Args:
            model: 推論セッション
            img_width (int): モデルのINPUTサイズ（input_dataが画像の場合は、画像の幅）
            img_height (int): モデルのINPUTサイズ（input_dataが画像の場合は、画像の高さ）
            input_data (Image | str): 推論するデータ（画像の場合RGB配列であること）
            labels (List[str], optional): クラスラベルのリスト. Defaults to None.
            colors (List[Tuple[int]], optional): ボックスの色のリスト. Defaults to None.
            nodraw (bool, optional): 描画フラグ. Defaults to False.

        Returns:
            Tuple[Dict[str, Any], Image]: 予測結果と出力画像(RGB)のタプル
        """
        # RGB画像をBGR画像に変換
        img_npy = convert.img2npy(input_data)
        img_npy = convert.bgr2rgb(img_npy)

        #input_shape = (img_width, img_height)
        #img, ratio = self.preprocess(img_npy, input_shape)

        #output = session.run(None, {session.get_inputs()[0].name: img[None, :, :, :]})
        from mmdet.apis import inference_detector
        result = inference_detector(model, img_npy)
        boxes = result.pred_instances.bboxes.cpu().numpy().tolist()
        boxes = [[row[1],row[0],row[3],row[2]] for row in boxes]
        ids = [i for i in range(len(boxes))]
        scores = result.pred_instances.scores.cpu().numpy().tolist()
        clses = result.pred_instances.labels.cpu().numpy().tolist()
        output_image, output_labels = common.draw_boxes(input_data, boxes, scores, clses, ids=ids, labels=labels, colors=colors, nodraw=nodraw)
        return dict(output_ids=ids, output_scores=scores, output_classes=clses, output_labels=output_labels, output_boxes=boxes), output_image
